/*

Copyright 2019 Brenton Leighton

This file is part of Mantle.

Mantle is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Mantle is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Mantle.  If not, see <https://www.gnu.org/licenses/>.

 */

const Gtk = imports.gi.Gtk;
const GObject = imports.gi.GObject;

const ExtensionUtils = imports.misc.extensionUtils;
const Me = ExtensionUtils.getCurrentExtension();
const Convenience = Me.imports.convenience;

let _settings;

// Key bindings
const _accel_names = [
    'tile-sw',
    'tile-s',
    'tile-se',
    'tile-w',
    'tile-maximised',
    'tile-e',
    'tile-nw',
    'tile-n',
    'tile-ne',
    'tile-centred',
    'focus-n',
    'focus-s',
    'focus-w',
    'focus-e',
    'focus-next-monitor',
    'move-next-monitor',
    'focus-last-window-current-monitor'];

function init() {
    _settings = Convenience.getSettings();
}

function _appendPreferencesPage(notebook) {

    let list_box = new Gtk.ListBox();

    // Split work area into thirds horizontally
    {
        let row = new Gtk.ListBoxRow();

        let box = new Gtk.HBox({margin_bottom: 10});

        let key = _settings.settings_schema.get_key('tiling-split-thirds');

        let label = new Gtk.Label({
            label: key.get_summary(),
            tooltip_text: key.get_description()
        });

        let light_switch = new Gtk.Switch({state: _settings.get_boolean('tiling-split-thirds')});

        light_switch.connect('state-set',
            function () {
                _settings.set_boolean('tiling-split-thirds', light_switch.get_active());
            });

        box.pack_start(label, false, false, 0);
        box.pack_end(light_switch, false, false, 0);

        row.add(box);
        list_box.add(row);
    }

    // Other tiling geometries
    {
        let row = new Gtk.ListBoxRow();

        let box = new Gtk.HBox({margin_bottom: 10});

        let key = _settings.settings_schema.get_key('tiling-centred-sizes');

        let label = new Gtk.Label({
            label: key.get_summary(),
            tooltip_text: key.get_description()
        });

        let entry = new Gtk.Entry({text: _settings.get_string('tiling-centred-sizes')});

        entry.connect('activate',
            function () {
                _settings.set_string('tiling-centred-sizes', entry.get_text());
            });

        box.pack_start(label, false, false, 0);
        box.pack_end(entry, false, false, 0);

        row.add(box);
        list_box.add(row);
    }

    // Merge other geometries into centre key
    {
        let options = ["Don't merge", "Merge before", "Merge after"];

        let row = new Gtk.ListBoxRow();

        let box = new Gtk.HBox({margin_bottom: 10});

        let key = _settings.settings_schema.get_key('tiling-merge-centred-maximised');

        let label = new Gtk.Label({
            label: key.get_summary(),
            tooltip_text: key.get_description()
        });

        let combo_box_text = new Gtk.ComboBoxText();

        for (let o of options) {
            combo_box_text.append_text(o);
        }

        combo_box_text.set_active(_settings.get_int('tiling-merge-centred-maximised'));

        combo_box_text.connect('changed',
            function () {
                _settings.set_int('tiling-merge-centred-maximised', combo_box_text.get_active());
            });

        box.pack_start(label, false, false, 0);
        box.pack_end(combo_box_text, false, false, 0);

        row.add(box);
        list_box.add(row);
    }

    // Merge bottom left and top left into left key, and bottom right and top right into right key
    {
        let row = new Gtk.ListBoxRow();

        let box = new Gtk.HBox({margin_bottom: 10});

        let key = _settings.settings_schema.get_key('tiling-merge-sides');

        let label = new Gtk.Label({
            label: key.get_summary(),
            tooltip_text: key.get_description()
        });

        let light_switch = new Gtk.Switch({state: _settings.get_boolean('tiling-merge-sides')});

        light_switch.connect('state-set',
            function () {
                _settings.set_boolean('tiling-merge-sides', light_switch.get_active());
            });

        box.pack_start(label, false, false, 0);
        box.pack_end(light_switch, false, false, 0);

        row.add(box);
        list_box.add(row);
    }

    // Tiling compare threshold
    {
        let row = new Gtk.ListBoxRow();

        let box = new Gtk.HBox({margin_bottom: 10});

        let key = _settings.settings_schema.get_key('tiling-compare-threshold');

        let label = new Gtk.Label({
            label: key.get_summary(),
            tooltip_text: key.get_description()
        });

        let entry = new Gtk.Entry({text: _settings.get_int('tiling-compare-threshold').toString()});

        entry.connect('activate',
            function () {
                _settings.set_int('tiling-compare-threshold', parseInt(entry.get_text()));
            });

        box.pack_start(label, false, false, 0);
        box.pack_end(entry, false, false, 0);

        row.add(box);
        list_box.add(row);
    }

    // Switch focus only on current monitor
    {
        let row = new Gtk.ListBoxRow();

        let box = new Gtk.HBox({margin_bottom: 10});

        let key = _settings.settings_schema.get_key('switch-focus-current-monitor-only');

        let label = new Gtk.Label({
            label: key.get_summary(),
            tooltip_text: key.get_description()
        });

        let light_switch = new Gtk.Switch({state: _settings.get_boolean('switch-focus-current-monitor-only')});

        light_switch.connect('state-set',
            function () {
                _settings.set_boolean('switch-focus-current-monitor-only', light_switch.get_active());
            });

        box.pack_start(label, false, false, 0);
        box.pack_end(light_switch, false, false, 0);

        row.add(box);
        list_box.add(row);
    }

    // Forward direction when switching focus of stacked windows
    {
        let options = ["Right", "Left"];

        let row = new Gtk.ListBoxRow();

        let box = new Gtk.Box();

        let key = _settings.settings_schema.get_key('switch-focus-forward-direction');

        let label = new Gtk.Label({
            label: key.get_summary(),
            tooltip_text: key.get_description()
        });

        let combo_box_text = new Gtk.ComboBoxText();

        for (let o of options) {
            combo_box_text.append_text(o);
        }

        combo_box_text.set_active(_settings.get_int('switch-focus-forward-direction'));

        combo_box_text.connect('changed',
            function () {
                _settings.set_int('switch-focus-forward-direction', combo_box_text.get_active());
            });

        box.pack_start(label, false, false, 0);
        box.pack_end(combo_box_text, false, false, 0);

        row.add(box);
        list_box.add(row);
    }

    let label = new Gtk.Label({label: "Preferences"});

    let box = new Gtk.VBox({
        can_focus: false,
        margin_top: 20,
        margin_bottom: 20,
        margin_left: 20,
        margin_right: 20,
        spacing: 10});

    box.pack_start(list_box, true, true, 0);

    let scrolled_window = new Gtk.ScrolledWindow({max_content_height: 350});
    scrolled_window.add(box);

    notebook.append_page(scrolled_window, label);

}

function _appendShortcutsPage(notebook) {

    let list_store = new Gtk.ListStore();

    // Name, summary, key value, modifier mask
    list_store.set_column_types([GObject.TYPE_STRING,
        GObject.TYPE_STRING,
        GObject.TYPE_STRING,
        GObject.TYPE_INT,
        GObject.TYPE_INT]);

    for (let i in _accel_names) {

        let name = _accel_names[i];

        let key = _settings.settings_schema.get_key(name);

        let accelerator = _settings.get_strv(name)[0];

        let key_value = 0;
        let modifier_mask = 0;

        if (accelerator != null) {
            [key_value, modifier_mask] = Gtk.accelerator_parse(accelerator);
        }

        list_store.set(list_store.append(),
            [0, 1, 2, 3, 4],
            [name, key.get_summary(), key.get_description(), key_value, modifier_mask]);

    }

    let tree_view = new Gtk.TreeView({
        expand: true,
        model: list_store});

    // First column
    let cell_renderer_text = new Gtk.CellRendererText();
    let col0 = new Gtk.TreeViewColumn({title: "Function"});
    col0.pack_start(cell_renderer_text, true);
    col0.add_attribute(cell_renderer_text, 'text', 1);
    tree_view.append_column(col0);

    // Second column
    let cell_renderer_accel = new Gtk.CellRendererAccel({
        'editable': true,
        'accel-mode': Gtk.CellRendererAccelMode.GTK
    });

    cell_renderer_accel.connect('accel-edited',
        function (renderer, path_string, accel_key, accel_mods) {

            // Get the item in the list store
            let [success, iter] = list_store.get_iter_from_string(path_string);

            if (!success) {
                throw new Error("Error getting iterator from list store");
            }

            // Set the accelerator values in the list store
            list_store.set(iter, [3, 4], [accel_key, accel_mods]);

            // Get the key name and accelerator name
            let name = list_store.get_value(iter, 0);
            let accel_name = Gtk.accelerator_name(accel_key, accel_mods);

            // Change the setting
            _settings.set_strv(name, [accel_name]);
        });

    cell_renderer_accel.connect('accel-cleared',
        function (renderer, path_string) {

            // Get the item in the list store
            let [success, iter] = list_store.get_iter_from_string(path_string);

            if (!success) {
                throw new Error("Error getting iterator from list store");
            }

            // Set the accelerator values in the list store
            list_store.set(iter, [3, 4], [0, 0]);

            // Get the key name and accelerator name
            let name = list_store.get_value(iter, 0);

            // Change the setting
            _settings.set_strv(name, null);
        });

    let col1 = new Gtk.TreeViewColumn({title: "Shortcut Key"});
    col1.pack_end(cell_renderer_accel, false);
    col1.add_attribute(cell_renderer_accel, 'accel-key', 3);
    col1.add_attribute(cell_renderer_accel, 'accel-mods', 4);
    tree_view.append_column(col1);

    let box = new Gtk.VBox({
        can_focus: false,
        margin_top: 20,
        margin_bottom: 20,
        margin_left: 20,
        margin_right: 20,
        spacing: 10});

    box.pack_start(tree_view, true, true, 0);

    // let scrolled_window = new Gtk.ScrolledWindow({max_content_height: 350});
    // scrolled_window.add(box);

    let label = new Gtk.Label({label: "Keyboard Shortcuts"});
    notebook.append_page(box, label);

}

function buildPrefsWidget() {

    let notebook = new Gtk.Notebook();

    _appendPreferencesPage(notebook);
    _appendShortcutsPage(notebook);

    notebook.show_all();

    return notebook;

}
