/*

Copyright 2019 Brenton Leighton

This file is part of Mantle.

Mantle is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Mantle is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Mantle.  If not, see <https://www.gnu.org/licenses/>.

 */

const Meta = imports.gi.Meta;

const ExtensionUtils = imports.misc.extensionUtils;
const Me = ExtensionUtils.getCurrentExtension();
const Convenience = Me.imports.convenience;

const _settings = Convenience.getSettings();

let _current_monitor_only = false;
let _left_forward = false;

let _previously_activated_window_id = [];
let _previous_window_stack_ids = [];

function _rectToCentre(rect) {
    return [(rect.x + rect.width / 2), (rect.y + rect.height / 2)];
}

function _getStackedWindows(focus_window_centre) {

    // Get all active workspace windows
    const active_workspace = global.screen.get_active_workspace();
    const all_windows = global.display.get_tab_list(Meta.TabList.NORMAL, active_workspace);

    let stacked_windows = [];

    for (let window of all_windows) {
        // Include the focused window
        if (window.has_focus()) {
            stacked_windows.push(window);
            continue;
        }

        const centre = _rectToCentre(window.get_frame_rect());

        if (centre[0] === focus_window_centre[0] && centre[1] === focus_window_centre[1]) {
            stacked_windows.push(window);
        }

    }

    return stacked_windows;
}

function _getIdArray(windows) {

    let ids = [];

    for (let window of windows) {
        ids.push(window.get_description().slice(0, 9));
    }

    return ids;
}

function _compareIdArrays(current, previous) {

    // If previous is missing anything from current (a new window has been added), return nothing
    // Otherwise return the previous array with anything not in the current array removed

    let previous_found = [];

    for (let p of previous) {
        previous_found.push(false);
    }

    for (let c of current) {
        let found = false;

        for (let j = 0; j < previous.length; j++) {
            if (!previous_found[j] && (c === previous[j])) {
                found = true;
                previous_found[j] = true;
                break;
            }
        }

        if (!found) {
            // Previous missing window from current, return empty array
            return [];
        }
    }

    // Return previous with anything missing from current removed
    let new_previous = [];

    for (let i = 0; i < previous.length; i++) {
        if (previous_found[i]) {
            new_previous.push(previous[i]);
        }
    }

    return new_previous;
}

function _getWindowsInDirection(focus_window_centre, direction) {

    // Get all active workspace windows
    const active_workspace = global.screen.get_active_workspace();
    const all_windows = global.display.get_tab_list(Meta.TabList.NORMAL, active_workspace);

    // Setting for switching focus on current monitor only
    // const current_monitor_only = _settings.get_boolean('switch-focus-current-monitor-only');
    const current_monitor = global.display.focus_window.get_monitor();

    let windows = [];

    for (let window of all_windows) {

        // Skip if the window is the focused window
        if (window.has_focus()) {
            continue;
        }

        // Skip if current_monitor_only is enabled, and the window is not on the current monitor
        if (_current_monitor_only && window.get_monitor() !== current_monitor) {
            continue;
        }

        // Get the centre of the window
        const centre = _rectToCentre(window.get_frame_rect());

        // Distances between this window and the focus window
        let diff_x = Math.abs(centre[0] - focus_window_centre[0]);
        let diff_y = Math.abs(centre[1] - focus_window_centre[1]);

        // Add the window to the array if it's in the correct direction
        switch (direction) {
            case 'n':
                if (centre[1] < focus_window_centre[1] && diff_y > (diff_x / 2)) {
                    windows.push(window);
                }
                break;
            case 's':
                if (centre[1] > focus_window_centre[1] && diff_y > (diff_x / 2)) {
                    windows.push(window);
                }
                break;
            case 'e':
                if (centre[0] > focus_window_centre[0] && diff_x > (diff_y / 2)) {
                    windows.push(window);
                }
                break;
            case 'w':
                if (centre[0] < focus_window_centre[0] && diff_x > (diff_y / 2)) {
                    windows.push(window);
                }
                break;
        }
    }

    return windows;
}

function _selectNearestWindow(focus_window_centre, direction, windows) {

    //Find the closest window
    let w = 0;
    let dist = 999999;

    for (let i = 0; i < windows.length; i++) {

        const rect = windows[i].get_frame_rect();
        const centre = _rectToCentre(rect);

        // Calculate distance
        let new_dist;
        if (direction === 'e' || direction === 'w') {
            new_dist = Math.abs(focus_window_centre[0] - centre[0]) +
                2 * Math.abs(focus_window_centre[1] - centre[1]);
        } else {
            new_dist = Math.abs(focus_window_centre[1] - centre[1])
                + 2 * Math.abs(focus_window_centre[0] - centre[0]);
        }

        // Update only if the window is more than 10 pixels closer
        // So for windows with similar distance the more recently used window is selected
        if (new_dist < dist - 10) {
            w = i;
            dist = new_dist;
        }
    }

    return w;

}

function _activateNearestWindow(focus_window_centre, direction) {
    // Windows in direction
    const windows = _getWindowsInDirection(focus_window_centre, direction);

    // If there are no windows, do nothing
    if (windows.length === 0) {
        return;
    }

    // If there's only one window, activate it
    if (windows.length === 1) {
        windows[0].activate(global.get_current_time());
        _previously_activated_window_id = [];
        _previous_window_stack_ids = [];
        return;
    }

    //Find closest window
    const w = _selectNearestWindow(focus_window_centre, direction, windows);

    windows[w].activate(global.get_current_time());
    _previously_activated_window_id = [];
    _previous_window_stack_ids = [];
}

function updateSettingCurrentMonitorOnly() {
    _current_monitor_only = _settings.get_boolean('switch-focus-current-monitor-only');
}

function updateSettingForwardDirection() {
    _left_forward = _settings.get_int('switch-focus-forward-direction') === 1;
}

// Focus the nearest window in a direction from the currently focused window
function focusWindow(direction) {

    const focus_window = global.display.focus_window;

    const focus_window_rect = focus_window.get_frame_rect();
    const focus_window_centre = _rectToCentre(focus_window_rect);

    // If the current focused window is not the previously activated window, reset
    const focus_window_id = focus_window.get_description().slice(0, 9);

    if (focus_window_id !== _previously_activated_window_id) {
        _previously_activated_window_id = [];
        _previous_window_stack_ids = [];
    }

    // Get windows in stack (including focus window)
    const stacked_windows = _getStackedWindows(focus_window_centre);

    if (stacked_windows.length > 1) {
        // Focused window is in a stack
        //global.log('Focused window is in a stack');

        if (_previous_window_stack_ids.length === 0) {
            // There is no previous stack
            //global.log('No previous stack');

            if (direction === 's' || (direction === 'e' && !_left_forward) || (direction === 'w' && _left_forward)) {
                // Activate the next window in the current stack
                //global.log('Activate next window in stack');

                stacked_windows[1].activate(global.get_current_time());

                // Save the activated window and stack ids
                _previously_activated_window_id = stacked_windows[1].get_description().slice(0, 9);
                _previous_window_stack_ids = _getIdArray(stacked_windows);

            } else {
                // Reverse direction, try to exit stack
                //global.log('Try to exit stack');

                _activateNearestWindow(focus_window_centre, direction);
            }

        } else {
            // There is a previous stack
            //global.log('There is a previous stack');

            let current_window_stack_ids = _getIdArray(stacked_windows);

            // Update the previous stack
            _previous_window_stack_ids = _compareIdArrays(current_window_stack_ids, _previous_window_stack_ids);

            if (_previous_window_stack_ids.length === 0) {
                // New window added
                //global.log('New window added to stack');

                if (direction === 's' || (direction === 'e' && !_left_forward) || (direction === 'w' && _left_forward)) {
                    // Activate the next window in the current stack
                    //global.log('Activate next window in stack');

                    stacked_windows[1].activate(global.get_current_time());

                    // Save the activated window and stack ids
                    _previously_activated_window_id = stacked_windows[1].get_description().slice(0, 9);
                    _previous_window_stack_ids = current_window_stack_ids;

                } else {
                    // Reverse direction, exit stack
                    //global.log('Try to exit stack');

                    _activateNearestWindow(focus_window_centre, direction);
                }
            } else {
                // Activate the window before or after the focus window
                //global.log('No new window added to stack');

                let focus_window_index = 0;

                for (let i = 0; i < _previous_window_stack_ids.length; i++) {
                    if (_previous_window_stack_ids[i] === focus_window_id) {
                        focus_window_index = i;
                        break;
                    }
                }

                let window_to_activate_id = [];

                if (direction === 's' || (direction === 'e' && !_left_forward) || (direction === 'w' && _left_forward)) {
                    if (focus_window_index === _previous_window_stack_ids.length - 1) {
                        //global.log('Try to exit stack');

                        _activateNearestWindow(focus_window_centre, direction);
                        return;
                    } else {
                        window_to_activate_id = _previous_window_stack_ids[focus_window_index + 1];
                    }
                } else {
                    if (focus_window_index === 0) {
                        //global.log('Try to exit stack');

                        _activateNearestWindow(focus_window_centre, direction);
                        return;
                    } else {
                        window_to_activate_id = _previous_window_stack_ids[focus_window_index - 1];
                    }
                }

                //global.log('Activating window ' + window_to_activate_id);

                // Find window_to_activate
                for (let i = 0; i < current_window_stack_ids.length; i++) {
                    if (window_to_activate_id === current_window_stack_ids[i]) {

                        //global.log('Found');

                        stacked_windows[i].activate(global.get_current_time());

                        // Save the activated window
                        _previously_activated_window_id = window_to_activate_id;
                    }
                }
            }

        }
    } else {
        // Not in stack
        //global.log('Focused window is not in a stack');

        _activateNearestWindow(focus_window_centre, direction);

        _previously_activated_window_id = [];
        _previous_window_stack_ids = [];
    }

}
