/*

Copyright 2019 Brenton Leighton

This file is part of Mantle.

Mantle is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Mantle is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Mantle.  If not, see <https://www.gnu.org/licenses/>.

 */

const Meta = imports.gi.Meta;

const ExtensionUtils = imports.misc.extensionUtils;
const Me = ExtensionUtils.getCurrentExtension();
const Convenience = Me.imports.convenience;

const _settings = Convenience.getSettings();

let _monitor_direction_tile = [];

let _compare_threshold = 10;

function _getOtherGeometries() {

    let string1 = _settings.get_string('tiling-centred-sizes');

    let string1_split = string1.split(',');

    let other_geometries = [];

    for (let string2 of string1_split) {
        let string2_split = string2.trim().split('x');

        let w = parseInt(string2_split[0].trim());
        let h = parseInt(string2_split[1].trim());

        other_geometries.push([w, h]);
    }

    return other_geometries;
}

function _generateTiles(work_area) {

    let split_thirds = _settings.get_boolean('tiling-split-thirds');

    // Determine splits
    const width_half = Math.round(work_area.width / 2);
    const width_third = Math.round(work_area.width / 3);
    const height_half = Math.round(work_area.height / 2);

    // Generate tiles for every direction
    let direction_tiles = [];

    // 0: other
    {
        let tiles = [];

        let other_sizes = _getOtherGeometries();

        for (let other_size of other_sizes) {
            let w = other_size[0];
            let h = other_size[1];

            if ((work_area.width > w) && (work_area.height > h)) {
                const x = work_area.x + Math.round((work_area.width - w) / 2);
                const y = work_area.y + Math.round((work_area.height - h) / 2);
                tiles.push([x, y, w, h]);
            }
        }

        direction_tiles.push(tiles);
    }

    // 1: bottom left
    {
        let tiles = [];

        // Bottom half, left half
        tiles.push([work_area.x, work_area.y + height_half, width_half, height_half]);

        if (split_thirds) {
            // Bottom half, left third
            tiles.push([work_area.x, work_area.y + height_half, width_third, height_half]);

            // Bottom half, left two thirds
            tiles.push([work_area.x, work_area.y + height_half, 2 * width_third, height_half]);
        }

        direction_tiles.push(tiles);
    }

    // 2: bottom
    {
        let tiles = [];

        // Bottom half horizontally maximised
        tiles.push([0, work_area.y + height_half, 0, height_half]);

        if (split_thirds) {
            // Bottom half middle third
            tiles.push([work_area.x + width_third, work_area.y + height_half, width_third, height_half]);
        }

        direction_tiles.push(tiles);
    }

    // 3: bottom right
    {
        let tiles = [];

        // Bottom half, right half
        tiles.push([work_area.x + width_half, work_area.y + height_half, width_half, height_half]);

        if (split_thirds) {
            // Bottom half, right third
            tiles.push([work_area.x + 2 * width_third, work_area.y + height_half, width_third, height_half]);

            // Bottom half, right two thirds
            tiles.push([work_area.x + width_third, work_area.y + height_half, 2 * width_third, height_half]);
        }

        direction_tiles.push(tiles);
    }

    // 4: left
    {
        let tiles = [];

        // Left half vertically maximised
        tiles.push([work_area.x, 0, width_half, 0]);

        if (split_thirds) {
            // Left third vertically maximised
            tiles.push([work_area.x, 0, width_third, 0]);

            // Left two thirds vertically maximised
            tiles.push([work_area.x, 0, 2 * width_third, 0]);
        }

        direction_tiles.push(tiles);
    }

    // 5: centre
    {
        let tiles = [];

        // Maximised
        tiles.push([0, 0, 0, 0]);

        if (split_thirds) {
            // Middle third vertically maximised
            tiles.push([work_area.x + width_third, 0, width_third, 0]);
        }

        direction_tiles.push(tiles);
    }

    // 6: right
    {
        let tiles = [];

        // Right half vertically maximised
        tiles.push([work_area.x + width_half, 0, width_half, 0]);

        if (split_thirds) {
            // Right third vertically maximised
            tiles.push([work_area.x + 2 * width_third, 0, width_third, 0]);

            // Right two thirds vertically maximised
            tiles.push([work_area.x + width_third, 0, 2 * width_third, 0]);
        }

        direction_tiles.push(tiles);
    }

    // 7: top left
    {
        let tiles = [];

        // Top half, left half
        tiles.push([work_area.x, work_area.y, width_half, height_half]);

        if (split_thirds) {
            // Top half, left third
            tiles.push([work_area.x, work_area.y, width_third, height_half]);

            // Top half, left two thirds
            tiles.push([work_area.x, work_area.y, 2 * width_third, height_half]);
        }

        direction_tiles.push(tiles);
    }

    // 8: top
    {
        let tiles = [];

        // Top half horizontally maximised
        tiles.push([0, work_area.y, 0, height_half]);

        if (split_thirds) {
            // Top half middle third
            tiles.push([work_area.x + width_third, work_area.y, width_third, height_half]);
        }

        direction_tiles.push(tiles);
    }

    // 9: top right
    {
        let tiles = [];

        // Top half, right half
        tiles.push([work_area.x + width_half, work_area.y, width_half, height_half]);

        if (split_thirds) {
            // Top half, right third
            tiles.push([work_area.x + 2 * width_third, work_area.y, width_third, height_half]);

            // Top half, right two thirds
            tiles.push([work_area.x + width_third, work_area.y, 2 * width_third, height_half]);
        }

        direction_tiles.push(tiles);
    }

    // Merge centred sizes into maximised
    switch (_settings.get_int('tiling-merge-centred-maximised')) {
        case 1:
            // Other before centre
            direction_tiles[5] = direction_tiles[0].concat(direction_tiles[5]);
            break;
        case 2:
            // Centre before other
            direction_tiles[5] = direction_tiles[5].concat(direction_tiles[0]);
            break;
    }

    // Merge sides
    if (_settings.get_boolean('tiling-merge-sides')) {
        direction_tiles[4] = direction_tiles[4].concat(direction_tiles[7], direction_tiles[1]);
        direction_tiles[6] = direction_tiles[6].concat(direction_tiles[9], direction_tiles[3]);
    }

    return direction_tiles;
}


function _selectTile(focus_window, tiles) {

    const focus_window_rect = focus_window.get_frame_rect();
    const focus_window_maximise_flags = focus_window.get_maximized();

    // Loop backwards
    for (let i = (tiles.length - 1); i >= 0; i--) {

        if (i === 0) {
            return tiles[0];
        }

        if (tiles[i - 1][2] === 0 && tiles[i - 1][3] === 0) {
            // If the previous tile is maximised
            if (focus_window_maximise_flags === Meta.MaximizeFlags.BOTH) {
                // And the focused window is maximised
                return tiles[i];
            }
        } else if (tiles[i - 1][3] === 0) {
            // If the previous tile is vertically maximised
            if (focus_window_maximise_flags === Meta.MaximizeFlags.VERTICAL &&
                (Math.abs(focus_window_rect.x - tiles[i - 1][0]) < _compare_threshold) &&
                (Math.abs(focus_window_rect.width - tiles[i - 1][2]) < _compare_threshold)) {
                // And the focused window is vertically maximised and the same horizontal position and size, select the current tile
                return tiles[i];
            }
        } else if (tiles[i - 1][2] === 0) {
            // If the previous tile is horizontally maximised
            if (focus_window_maximise_flags === Meta.MaximizeFlags.HORIZONTAL &&
                (Math.abs(focus_window_rect.y - tiles[i - 1][1]) < _compare_threshold) &&
                (Math.abs(focus_window_rect.height - tiles[i - 1][3]) < _compare_threshold)) {
                // And the focused window is horizontally maximised and the same vertical position and size, select the current tile
                return tiles[i]
            }
        } else {
            if ((Math.abs(focus_window_rect.x - tiles[i - 1][0]) < _compare_threshold) &&
                (Math.abs(focus_window_rect.y - tiles[i - 1][1]) < _compare_threshold) &&
                (Math.abs(focus_window_rect.width - tiles[i - 1][2]) < _compare_threshold) &&
                (Math.abs(focus_window_rect.height - tiles[i - 1][3]) < _compare_threshold)) {
                // If focused window has the same geometry (within the compare threshold) as the previous tile, select the current tile
                return tiles[i];
            }
        }
    }
}

function _setWindowGeometry(focus_window, tile) {

    focus_window.unmake_fullscreen();

    if (tile[2] === 0 && tile[3] === 0) {
        // Maximised tile
        focus_window.maximize(Meta.MaximizeFlags.BOTH);
        return;
    }

    if (tile[2] !== 0 && tile[3] !== 0) {
        // Unmaximised tile
        focus_window.unmaximize(Meta.MaximizeFlags.BOTH);
        focus_window.move_resize_frame(true, tile[0], tile[1], tile[2], tile[3]);

    } else if (tile[2] !== 0 && tile[3] === 0) {
        // Vertically maximised tile
        focus_window.unmaximize(Meta.MaximizeFlags.HORIZONTAL);
        focus_window.maximize(Meta.MaximizeFlags.VERTICAL);

        const focus_window_rect = focus_window.get_frame_rect();

        focus_window.move_resize_frame(true, tile[0], focus_window_rect.y, tile[2], focus_window_rect.height);

    } else if (tile[2] === 0 && tile[3] !== 0) {
        // Horizontally maximised tile
        focus_window.unmaximize(Meta.MaximizeFlags.VERTICAL);
        focus_window.maximize(Meta.MaximizeFlags.HORIZONTAL);

        const focus_window_rect = focus_window.get_frame_rect();

        focus_window.move_resize_frame(true, focus_window_rect.x, tile[1], focus_window_rect.width, tile[3]);

    }
}

function updateSettingCompareThreshold() {
    _compare_threshold = _settings.get_int('tiling-compare-threshold');
}

function init() {

    const n_monitors = global.screen.get_n_monitors();

    let workspace = global.screen.get_workspace_by_index(0);

    // Generate tiles for each direction and monitor
    _monitor_direction_tile = [];

    for (let i = 0; i < n_monitors; i++) {
        let work_area = workspace.get_work_area_for_monitor(i);

        _monitor_direction_tile.push(_generateTiles(work_area));
    }

}


// Cycle focused window through the tiled geometries for a given direction
function tileWindow(direction) {

    let focus_window = global.display.focus_window;
    let monitor = focus_window.get_monitor();

    let tile = _selectTile(focus_window, _monitor_direction_tile[monitor][direction]);

    _setWindowGeometry(focus_window, tile);

}
